#!/usr/bin/env python

from random import randint
from time import time
from colormap import colormap
from math import cos, pi, sin, sqrt

import os.path
import gtk, gtk.gdk, gtk.glade, pango

import plugins
import about, background

class Node:
    def __init__(self, id=-1, angle=0.0, radius=0.0, text=""):
        self.id          = id
        self.pos_angle   = angle
        self.pos_radius  = radius
        self.goal_angle  = angle
        self.goal_radius = radius
        self.angle_inc   = 0
        self.radius_inc  = 0
        self.text        = text
        self.visited     = 0

class Circle:
    def __init__(self):
        self.radius = 1
        self.nodes  = []

class Controller:
    def __init__(self):
        data = self.default_data()
        
        self.node_table = data[0]
        self.circle_list = []
        self.selected = data[1]
        self.timeout = 0.5
        self.tlast  = time()
        self.ttimeout = None
        self.idle_handler = -1      

        self.diameter = 100
        self.fisheye = 0.5
        self.animating = 0
        
        l_len = 0
        for l in self.node_table.keys():
            l_len = max(len(self.node_table[l]), l_len)

        self.circles = l_len + 1
        for i in range(self.circles):
            self.circle_list.append(Circle())

        self.plugins = []
        for plugin in plugins.__all__:
            __import__('plugins.' + plugin + '.setup')
            self.plugins.append(plugin)        

        self.window = GUI(self)

        self.refresh_data()      
        self.window.show()

    def default_data(self):
        n1 = Node(1)
        n2 = Node(2)
        n3 = Node(3)
        n4 = Node(4)
        n5 = Node(5)
        n6 = Node(6)
        n7 = Node(7)
        n8 = Node(8)
        n9 = Node(9)

        n1.text = "time"
        n2.text = "is"
        n3.text = "the"
        n4.text = "fire"
        n5.text = "in"
        n6.text = "which"
        n7.text = "we"
        n8.text = "burn"
        n9.text = "!"

        node_table = {n1 : [n2, n3, n4],
                      n2 : [n1],
                      n3 : [n1,n5,n6],
                      n4 : [n1],
                      n5 : [n3],
                      n6 : [n3,n7,n8],
                      n7 : [n6,n8,n9],
                      n8 : [n6,n7],
                      n9 : [n7]}

        return (node_table, n1)

    def refresh_data(self):
        # remove double dependencies in the graph
        # e.g. (n7->n6, n8->n7, n8->n6) would remove n8->n6 as
        # n8 is a direct descendant of n7
        for key in self.node_table.keys():
            for n in self.node_table[self.node_table[key][0]][1:]:
                if n != key:
                    if key in self.node_table[n]:
                        self.node_table[key].remove(n)
                        self.node_table[n].remove(key)

        self.recalc_circle_radii()

    def recalc_circle_radii(self):
        for i in range(1,len(self.circle_list),1):
            self.circle_list[i].radius = (self.circle_list[i-1].radius + self.diameter) * self.fisheye
            
        for c in self.circle_list:
            for n in c.nodes: n.visited = 0
            c.nodes = []
        
        self.reorder(self.selected)
        self.tlast = time()     

        if not self.animating:
            self.idle_handler = gtk.idle_add(self.window._update)
            self.animating = 1

    def reorder(self, selected, circle_nr=0, lower_angle=0, upper_angle=360, offset=0):
        if circle_nr == len(self.circle_list):
            return

        self.circle_list[circle_nr].nodes.append(selected)

        selected.goal_angle = offset + ((upper_angle - lower_angle) / 2)
        selected.goal_radius = self.circle_list[circle_nr].radius    

        selected.angle_inc  = (selected.goal_angle - selected.pos_angle)   / self.timeout
        selected.radius_inc = (selected.goal_radius - selected.pos_radius) / self.timeout
        self.ttimeout = time()+self.timeout
        selected.visited = 1

        node_cnt = len([n for n in self.node_table[selected] if not n.visited])
        if not node_cnt: return
        angle = (upper_angle - lower_angle) / node_cnt

        for n in self.node_table[selected]:
            if not n.visited:
                self.reorder(n, circle_nr + 1, lower_angle, lower_angle + angle, lower_angle)
                lower_angle += angle        
                


class GUI:
    LIBGLADE_FILE = "infoviz.glade"
    COLORMAP_ENTRIES = 50.0
    
    def __init__(self, controller):
        self.controller = controller

        xml = gtk.glade.XML(self.LIBGLADE_FILE)
        self.window = xml.get_widget('window_main')
        self.darea = xml.get_widget('drawingarea1')
        self.hpaned = xml.get_widget('hpaned1')
        self.label_speed = xml.get_widget('label_scroll_speed')
        self.label_circles = xml.get_widget('label_scroll_circles')
        self.label_diameter = xml.get_widget('label_scroll_diameter')        
        self.label_fisheye = xml.get_widget('label_scroll_fisheye')
        self.label_gradient = xml.get_widget('label_scroll_gradient')
        self.label_thickness = xml.get_widget('label_scroll_thickness')
        self.hscale_circles = xml.get_widget('hscale_circles')        
        self.hscale_gradient = xml.get_widget('hscale_gradient')
        self.plugin_area = xml.get_widget('plugin_area')
        self.menu_plugins = xml.get_widget('menu_plugins_menu')
        self.statusbar = xml.get_widget('statusbar')

        # initalize plugins
        self.menu_plugins.remove(xml.get_widget('empty'))
        for plugin in controller.plugins:
            menu = gtk.MenuItem(eval('plugins.'+plugin+'._PLUGIN_NAME'))
            menu.set_name(plugin)
            self.menu_plugins.append(menu)
            menu.connect('activate', self._on_menuplugin_activate)            
            menu.show()

        self.circles = self.controller.circles
        self.controller.diameter = 100
        self.controller.fisheye = 0.5
        self.gradient = self.controller.circles
        self.thickness = 10

        self.hscale_circles.set_value(self.circles)
        self.hscale_gradient.set_value(self.circles)
        self.hscale_gradient.set_range(1, self.circles)

        # flags
        self.demo_mode = 0
        self.background_off = 0
        self.show_about = 0
        self.plugin_loaded = ""

        xml.signal_autoconnect({
            'on_exit1_activate'                 : self._on_exit1_activate,
            'on_window_destroy_event'           : self._on_exit1_activate,
            'on_drawingarea_configure_event'    : self._on_drawingarea_configure_event,
            'on_window1_size_request'           : self._on_window1_size_request,
            'on_drawingarea_expose_event'       : self._on_drawingarea_expose_event,
            'on_drawingarea_button_press_event' : self._on_drawingarea_button_press_event,
            'on_about1_activate'                : self._on_about_activate,
            'on_hscale_speed_value_changed'     : self._on_hscale_speed_value_changed,
            'on_hscale_circles_value_changed'   : self._on_hscale_circles_value_changed,
            'on_hscale_diameter_value_changed'  : self._on_hscale_diameter_value_changed,
            'on_hscale_fisheye_value_changed'   : self._on_hscale_fisheye_value_changed,
            'on_hscale_gradient_value_changed'  : self._on_hscale_gradient_value_changed,
            'on_hscale_thickness_value_changed' : self._on_hscale_thickness_value_changed,
            'on_checkbuttondemomode_toggled'    : self._on_checkbuttondemomode_toggled,
            'on_checkbuttonbackground_toggled'  : self._on_checkbuttonbackground_toggled,            
            'on_button_font_clicked'            : self._on_button_font_clicked
            })
        
        (self.pixmap_about, mask) = gtk.gdk.pixmap_colormap_create_from_xpm_d(None, self.window.get_colormap(),
                                                                              None, about.about_xpm, )
        (self.pixmap_background, mask) = gtk.gdk.pixmap_colormap_create_from_xpm_d(None, self.window.get_colormap(),
                                                                                   None, background.background_xpm, )

        self.font = pango.FontDescription('Serif 12')
        # contextus id-us illuminatus *eg*
        self.statusbar.push(23, '')        
        
        # self.darea.window.set_back_pixmap(pixmap_background, 0)

        # make sure that configure is really called once before everything else
        self._on_drawingarea_configure_event(self.darea, None)

        self.plugin_widget = None
        # self.save_image_cnt = 0


    def show(self):
        gtk.main()
    
    def _on_exit1_activate(self, obj, obj2 = None):
        gtk.main_quit()

    def _on_menuplugin_activate(self, obj):
        if self.plugin_loaded != obj.get_name():
            if self.plugin_widget != None:
                self.plugin_area.remove(self.plugin_widget)
            self.plugin_loaded = obj.get_name()
            self.plugin_widget = apply(eval('plugins.' + obj.get_name() + '.setup.init'),
                                       (self.plugin_area, self.controller)).get_widget()

    def _deactivate_about_cb(self):
        self.show_about = 0
        self.statusbar.pop(23)
        self.draw()

    def _on_about_activate(self, obj, obj2 = None):
        self.show_about = 1
        self.statusbar.push(23, '.:InfooViz 0.2:. Programming by: Tamer "tuhtah" Fahmy, Gabriel "gabe" Wurzer, Matthias "grumble" Grumet')        
        gtk.timeout_add(10000, self._deactivate_about_cb)
        self.draw()

    def _on_drawingarea_configure_event(self, darea, event):
        self.win_width = self.darea.get_allocation()[2]
        self.win_height = self.darea.get_allocation()[3]
        
        self.pixmap = gtk.gdk.Pixmap(self.darea.window,
                                     self.win_width,
                                     self.win_height)        
        self.draw()
        return gtk.TRUE

    def _on_drawingarea_expose_event(self, darea, event):
        x, y, width, height = event.area
        darea.window.draw_drawable(darea.get_style().black_gc, self.pixmap,
                                   x, y, x, y, width, height)
        return gtk.FALSE        
        
    def _on_window1_size_request(self, window, requisition):
        self.hpaned.set_position(self.window.get_allocation()[2]-150)

    def _on_drawingarea_button_press_event(self, darea, event):
        if event.button == 1:
            for c in self.controller.circle_list:
                for n in c.nodes:
                    xn = self.win_width/2 + n.pos_radius * cos(n.pos_angle*pi*2/360)
                    yn = self.win_height/2 + n.pos_radius * sin(n.pos_angle*pi*2/360)

                    # compute vector n to click 
                    xv = event.x - xn
                    yv = event.y - yn

                    if sqrt(xv*xv+yv*yv) < self.thickness:
                        self.controller.selected = n
                        self.controller.recalc_circle_radii()                        

    def _on_hscale_speed_value_changed(self, orange):
        self.controller.timeout = orange.get_value()

    def _on_hscale_circles_value_changed(self, orange):
        if self.circles < orange.get_value():
            for i in range(int(orange.get_value()) - self.circles):
                self.controller.circle_list.append(Circle())
            self.circles = int(orange.get_value())
            self.hscale_gradient.set_sensitive(gtk.TRUE)                    
            self.hscale_gradient.set_range(1, self.circles)
        else:
            for circle in self.controller.circle_list[int(orange.get_value()):]:
                for n in circle.nodes:
                    n.visited = 0
            self.circles = int(orange.get_value())                    
            self.controller.circle_list = self.controller.circle_list[0:self.circles] 
            if self.gradient >= self.circles:
                self.gradient = self.circles
                self.hscale_gradient.set_value(self.circles)
                if self.circles > 1:
                    self.hscale_gradient.set_sensitive(gtk.TRUE)                    
                    self.hscale_gradient.set_range(1, self.circles)
                else:
                    self.hscale_gradient.set_sensitive(gtk.FALSE)
        
        self.controller.recalc_circle_radii()

    def _on_hscale_diameter_value_changed(self, orange):
        self.controller.diameter = orange.get_value()
        self.controller.recalc_circle_radii()

    def _on_hscale_fisheye_value_changed(self, orange):
        self.controller.fisheye = orange.get_value()
        self.controller.recalc_circle_radii()
        
    def _on_hscale_gradient_value_changed(self, orange):
        if orange.get_value() <= self.circles:
            self.gradient = int(orange.get_value())
            self.draw()
            
    def _on_hscale_thickness_value_changed(self, orange):
        self.thickness = orange.get_value()
        self.draw()

    def _on_checkbuttondemomode_toggled(self, toggled):
        self.demo_mode = toggled.get_active()
        if self.demo_mode:
            # self.save_image_cnt = 0
            self._random_select()

    def _on_checkbuttonbackground_toggled(self, toggled):
        self.background_off = toggled.get_active()
        self.draw()
    
    def _on_button_font_clicked(self, button):
        def on_ok(button, self, fsel):
            self.font = pango.FontDescription(fsel.get_font_name())
            fsel.destroy()
            self.draw()
            
        def on_cancel(button, fsel):
            fsel.destroy()
        
        fsel = gtk.FontSelectionDialog("Select Font")
        fsel.set_modal(gtk.TRUE)
        fsel.set_font_name(self.font.to_string())
        fsel.ok_button.connect("clicked", on_ok, self, fsel)
        fsel.cancel_button.connect("clicked", on_cancel, fsel)
        fsel.show()
    
    def _random_select(self):
        self.controller.selected = self.controller.node_table.keys()[randint(0, len(self.controller.node_table.keys())-1)]
        self.controller.recalc_circle_radii()
        
    def draw(self):
        drawable = self.pixmap

        gc = drawable.new_gc()
        cmap = drawable.get_colormap()

        if self.background_off:
            drawable.draw_rectangle(self.darea.get_style().black_gc, 1,
                                    0,0, self.win_width, self.win_height)
        else:
            # FIXME: dehardcode it! 20030620 tamer.
            # draw the tiled background
            for x in range(0, self.win_width, 192):
                for y in range(0, self.win_height, 191):
                    drawable.draw_drawable(self.darea.get_style().black_gc,
                                           self.pixmap_background,
                                           0, 0, x, y, -1, -1)
        
        if self.show_about:
            # FIXME: dehardcode it! 20030617 tamer.            
            drawable.draw_drawable(gc, self.pixmap_about, 0, 0,
                                   self.win_width/2 - 250, self.win_height/2 - 187, -1, -1)

        factor = (self.COLORMAP_ENTRIES / self.controller.circle_list[self.gradient-1].radius)

        for c in self.controller.circle_list:
            color = min(self.COLORMAP_ENTRIES-1,
                        max(factor * c.radius - 1, 0))
            gc.foreground = cmap.alloc_color(colormap[int(color)])
            try:
                drawable.draw_arc(gc, 0,
                                  self.win_width/2 - c.radius, self.win_height/2 - c.radius,
                                  c.radius*2, c.radius*2,
                                  0,360*64)
            except OverflowError:
                pass

            for n in c.nodes:
                # translation from polar to x,y coordinates
                radius = n.pos_radius
                x = self.win_width/2 + radius * cos(n.pos_angle*pi*2/360)
                y = self.win_height/2 + radius * sin(n.pos_angle*pi*2/360)

                gc.foreground = cmap.alloc_color(colormap[-1])

                for n2 in self.controller.node_table[n]:
                    if n2.visited:
                        try:
                            drawable.draw_line(gc, x, y,
                                               self.win_width/2 + n2.pos_radius * cos(n2.pos_angle*pi*2/360),
                                               self.win_height/2 + n2.pos_radius * sin(n2.pos_angle*pi*2/360))
                        except OverflowError:
                            pass

                color = min(self.COLORMAP_ENTRIES-1,
                            max(factor * n.pos_radius - 1, 0))
                gc.foreground = cmap.alloc_color(colormap[int(color)])

                try:
                    drawable.draw_arc(gc, 1,
                                      x-self.thickness/2, y-self.thickness/2,
                                      self.thickness,self.thickness,
                                      0,360*64)
                except OverflowError:
                    pass
                
                layout = self.darea.create_pango_layout(`n.id`)
                layout.set_font_description(self.font)
                text_width, text_height = layout.get_pixel_size()
                drawable.draw_layout(gc,
                                     x-text_width-self.thickness/3,
                                     y-text_height-self.thickness/3, layout)
                
                layout = self.darea.create_pango_layout(n.text)
                layout.set_font_description(self.font)
                drawable.draw_layout(gc,
                                     x+self.thickness/2,
                                     y+self.thickness/2, layout)

        self.darea.queue_draw()

        # code snippet to create mpeg movies
        # if 0 and self.demo_mode and self.save_image_cnt < 100:
        #     error = {}
        #     gtk.gdk.Pixbuf(gtk.gdk.COLORSPACE_RGB, 0, 8,
        #                    self.win_width,
        #                    self.win_height).get_from_drawable(drawable,cmap,
        #                                                       0,0,0,0,
        #                                                       self.win_width,
        #                                                       self.win_height).save('shots/img_%04d.png' % self.save_image_cnt,
        #                                                                             'png',
        #                                                                             error)
        #     self.save_image_cnt += 1
        
    def _update(self):
        t = time()         
        dt = t - self.controller.tlast
        self.controller.tlast = t
        
        for c in self.controller.circle_list:
            for n in c.nodes:
                if n.angle_inc:
                    n.pos_angle += dt * n.angle_inc
                    if t >= self.controller.ttimeout:
                        n.angle_inc  = 0
                        n.pos_angle = n.goal_angle
                        gtk.idle_remove(self.controller.idle_handler)
                        self.controller.animating = 0                        
                        if self.demo_mode:
                            self._random_select()
                        
                if n.radius_inc:
                    n.pos_radius += dt * n.radius_inc
                    if t >= self.controller.ttimeout:
                        n.radius_inc = 0
                        n.pos_radius = n.goal_radius
                        gtk.idle_remove(self.controller.idle_handler)
                        self.controller.animating = 0
                        if self.demo_mode:
                            self._random_select()

        self.draw()

        return gtk.TRUE

def main():
    Controller()
    
if __name__ == "__main__":
    main()
